using System;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using SimpleJSON;
using UnityEngine.UI;
using System.Diagnostics;
using System.CodeDom;

public class ActionListManager : Persistence
{
    private int newActionCounter = 1;
    public bool lockedOnTransitionAction = false;
    private JSONStorableStringChooser actionsChooser;

    private ActionTextFieldsUI actionTextFieldsUI;

    private LinkedActionsUI linkedActionsUI;

    private NewLinkedActionsUI newLinkedActionsUI;

    private ActionItem deferredActionItem;

    private ActionItem lastAction;

    private List<Trigger> activeTriggers = new List<Trigger>();

    private List<UIDynamic> uiElements = new List<UIDynamic>();

    private bool scheduleRefreshUI;

    private UnityEventsListener PluginUIEventsListener { get; set; }

    public override void Init()
    {
        try
        {
            InitStorables();
            InitPluginActions();
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(ActionListManager)}.{nameof(Init)}: {e}");
        }
    }
    public override void InitUI()
    {
        base.InitUI();
        if (!UITransform || PluginUIEventsListener)
        {
            return;
        }

        PluginUIEventsListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
        PluginUIEventsListener.EnabledHandlers += () =>
        {
            if (scheduleRefreshUI)
            {
                scheduleRefreshUI = false;
                RefreshUI();
            }
        };

        InitOptionsUI();
        RefreshUI();

        //set background color
        UITransform.GetComponentInChildren<MVRScriptUI>().rightUIContent.parent.parent.parent.transform.GetComponent<Image>().color = Colors.MAIN_WINDOW;
        UITransform.GetComponentInChildren<MVRScriptUI>().rightUIContent.parent.parent.parent.parent.transform.GetComponent<Image>().color = Colors.MAIN_WINDOW_BACK;
    }

    public void RefreshActionUI()
    {
        if (!ShouldDeferRefresh() && actionsDictionary.Count() > 0)
        {
            actionTextFieldsUI.RefreshUI();
            linkedActionsUI.RefreshUI();
            newLinkedActionsUI.RefreshUI();
        }
    }

    private void InitOptionsUI()
    {
        Utils.OnInitUI(CreateUIElement);
        actionsChooser = new JSONStorableStringChooser("Actions", actionsDictionary.Keys.ToList(), "", "Action:", OnActionSelected);

        linkedActionsUI = new LinkedActionsUI(this);
        actionTextFieldsUI = new ActionTextFieldsUI(this);
        linkedActionsUI.Init(actionTextFieldsUI, actionsChooser, actionsDictionary);
        actionTextFieldsUI.Init(linkedActionsUI, actionsChooser, actionsDictionary);
        newLinkedActionsUI = new NewLinkedActionsUI(this, actionsChooser, actionsDictionary, linkedActionsUI);
        newLinkedActionsUI.InitUI();
    }

    private void OnExecuteActionButtonClicked()
    {
        OnTriggerSelectedActionChange(actionsChooser.val);
        OnTriggerStateChange();
    }

    private void OnDuplicateActionButtonClicked()
    {
        string actionToDuplicate = actionsChooser.val;
        if (string.IsNullOrEmpty(actionToDuplicate))
        {
            return;
        }

        if (actionsDictionary.ContainsKey(actionToDuplicate))
        {
            ActionItem duplicatedAction = new ActionItem(GetUniqueDuplicateName(actionToDuplicate), actionsDictionary[actionToDuplicate]);
            actionsDictionary.Add(duplicatedAction.Name, duplicatedAction);

            actionsChooser.val = duplicatedAction.Name;
            linkedActionsUI.RefreshUI();
        }
        else
        {
            SuperController.LogError($"Action '{actionToDuplicate}' not found in dictionary.");
        }
    }

    private string GetUniqueDuplicateName(string baseName)
    {
        int counter = 1;
        string newName;
        do
        {
            newName = $"{baseName} Copy_{counter++}";
        } while (actionsDictionary.ContainsKey(newName));
        return newName;
    }

    private void OnRemoveActionButtonClicked()
    {
        string actionToRemove = actionsChooser.val;
        if (string.IsNullOrEmpty(actionToRemove))
        {
            return;
        }

        if (actionsDictionary.ContainsKey(actionToRemove))
        {
            foreach (var key in new List<string>(actionsDictionary.Keys))
            {
                actionsDictionary[key].LinkedActions.Remove(actionToRemove);
            }

            int indexToRemove = actionsChooser.choices.IndexOf(actionToRemove);
            int nextIndex = 0;
            if (indexToRemove >= 0 && indexToRemove < actionsChooser.choices.Count - 1)
            {
                nextIndex = indexToRemove;
            }

            actionsDictionary.Remove(actionToRemove);

            actionsChooser.choices = actionsDictionary.Keys.ToList();

            if (actionsChooser.choices.Count > 0)
            {
                actionsChooser.val = actionsChooser.choices[nextIndex];
            }
            else
            {
                actionTextFieldsUI.resetAllFields();
                linkedActionsUI.resetAllFields();
                actionsChooser.val = "";
                if (actionsDictionary.Count() == 0)
                {
                    actionTextFieldsUI.RemoveAllUI();
                    linkedActionsUI.RemoveAllUI();
                }
                return;
            }

            RefreshUI();
        }
        else
        {
            SuperController.LogError("Action to remove not found in dictionary.");
        }
    }

    private void OnActionSelected(string selectedAction)
    {
        if (string.IsNullOrEmpty(selectedAction))
        {
            return;
        }

        ActionItem actionItem;
        if (actionsDictionary.TryGetValue(selectedAction, out actionItem))
        {
            actionTextFieldsUI.copyFieldValues(actionItem);
            linkedActionsUI.copyFieldValues(actionItem);

            linkedActionsUI.sameActionNameFunction.valNoCallback = ActionNameUtils.FunctionFormatActionName(actionItem.Name) == actionItem.Action ? true : false;

            linkedActionsUI.showAdvancedSettings = false;

            RefreshActionUI();
        }
    }

    private void OnAddActionButtonClicked()
    {
        string nameValue = "new action " + newActionCounter++;
        string actionValue = ActionNameUtils.FunctionFormatActionName(nameValue);
        string whenValue = "<type_in_trigger_condition>";

        ActionItem newActionItem = new ActionItem(nameValue, actionValue, whenValue, this);

        actionsDictionary.Add(actionValue, newActionItem);
        actionsChooser.choices = actionsDictionary.Keys.ToList();
        actionsChooser.val = actionValue;

        RefreshActionUI();
    }

    private void InitPluginActions()
    {
        RegisterAction(new JSONStorableAction("TriggerActionChange", OnTriggerActionChange));
        RegisterAction(new JSONStorableAction("TriggerStateChange", OnTriggerStateChange));
        RegisterAction(new JSONStorableAction("TransitionActionFinished", OnTransitionActionFinished));
    }

    private void InitStorables()
    {
        RegisterString(new JSONStorableString("TriggerSelectedActionChange", "", OnTriggerSelectedActionChange));
    }

    private void OnTransitionActionFinished()
    {
        lockedOnTransitionAction = false;
        UpdateActionList(lastAction);
        UpdateCurrentContext(lastAction);
        RefreshUI();
        actionsChooser.val = lastAction.Name;
    }
    private void OnTriggerStateChange()
    {
        string newState = getPluginByType(containingAtom, "Voxta").GetStringChooserJSONParam("State").val;
        if ((newState == "idle" || newState == "off") && deferredActionItem != null)
        {
            ExecuteAction(deferredActionItem);
            SetChooserToNewActionWhenLinkedToOthers(deferredActionItem);
            deferredActionItem = null;
        }
    }

    private void OnTriggerSelectedActionChange(string name)
    {
        if (actionsDictionary.Count > 0)
        {
            ExecuteAction(GetItemByName(name));
        }
    }

    private void OnTriggerActionChange()
    {
        string newAction = getPluginByType(containingAtom, "Voxta").GetStringJSONParam("CurrentAction").val;
        if (!string.IsNullOrEmpty(newAction))
        {
            ActionItem actionItem = FindActionRoute(newAction);
            if (actionItem.Deferred == false)
            {
                ExecuteAction(actionItem);
                SetChooserToNewActionWhenLinkedToOthers(actionItem);
            }
            else
            {
                deferredActionItem = actionItem;
            }
        }
    }

    private ActionItem FindActionRoute(string action)
    {
        foreach (string name in lastAction.LinkedActions)
        {
            ActionItem actionItem = GetItemByName(name);
            if (actionItem.Action == action)
            {
                return actionItem;
            }
        }

        SuperController.LogError("Did not find route for action = " + action + " in lastAction with name = " + lastAction.Name);
        return null;
    }

    private void SetChooserToNewActionWhenLinkedToOthers(ActionItem actionItem)
    {
        if (actionItem.LinkedActions.Count > 0)
        {
            actionsChooser.val = actionItem.Name;
        }
    }

    public void ExecuteAction(ActionItem actionItem)
    {
        actionItem.Trigger.Trigger();

        if (actionItem.Trigger.HasActionsWithActiveTimers)
        {
            activeTriggers.Add(actionItem.Trigger);
        }

        if (actionItem.PlayTimelineAnim)
        {
            var timelinePlugin = getPluginByType(containingAtom, "VamTimeline.AtomPlugin", true);
            if (timelinePlugin != null)
            {
                timelinePlugin.CallAction("Play VA_" + actionItem.Action + "/*");
                timelinePlugin.CallAction("Play VA_" + actionItem.Action);
                if (actionItem.TransitionAction)
                {
                    lockedOnTransitionAction = true;
                    actionsChooser.val = actionItem.Name;
                    UpdateActionList(actionItem);
                    RefreshUI();
                    return;
                }
            }
        }

        UpdateActionList(actionItem);
        UpdateCurrentContext(actionItem);
        RefreshActionUI();
    }

    private void UpdateActionList(ActionItem actionItem)
    {
        var newActionsList = GetActionAndLinkedActionsAsFormattedString(actionItem);
        if (!string.IsNullOrEmpty(newActionsList) || actionItem.TransitionAction)
        {
            getPluginByType(containingAtom, "Voxta").GetStringJSONParam("ActionsList").val = newActionsList;
            lastAction = actionItem;
        }
    }

    private void UpdateCurrentContext(ActionItem actionItem)
    {
        if (!string.IsNullOrEmpty(actionItem.Context))
        {
            currentContext.FreeText = actionItem.Context;
        }
        if (!string.IsNullOrEmpty(actionItem.Location))
        {
            currentContext.Location = actionItem.Location;
        }
        if (!string.IsNullOrEmpty(actionItem.Wearing))
        {
            currentContext.Wearing = actionItem.Wearing;
        }

        // Build context string for Voxta
        var contextBuilder = new System.Text.StringBuilder();
        if (!string.IsNullOrEmpty(currentContext.FreeText))
        {
            contextBuilder.Append(currentContext.FreeText).Append(". ");
        }
        if (!string.IsNullOrEmpty(currentContext.Location))
        {
            contextBuilder.Append("{{char}} is in ").Append(currentContext.Location).Append(". ");
        }
        if (currentContext.Wearing == "nothing")
        {
            contextBuilder.Append("{{char}} is naked.");
        }
        else if (!string.IsNullOrEmpty(currentContext.Wearing))
        {
            contextBuilder.Append("{{char}} is wearing ").Append(currentContext.Wearing).Append(".");
        }

        //SuperController.LogError("Updating context: " + contextBuilder.ToString());

        getPluginByType(containingAtom, "Voxta").GetStringJSONParam("Context").val = contextBuilder.ToString();
    }

    private string GetActionAndLinkedActionsAsFormattedString(ActionItem actionItem)
    {
        var stringBuilder = new System.Text.StringBuilder();

        foreach (var linkedActionKey in actionItem.LinkedActions)
        {
            ActionItem linkedActionItem;
            if (actionsDictionary.TryGetValue(linkedActionKey, out linkedActionItem))
            {
                if (actionItem.TransitionAction && string.IsNullOrEmpty(linkedActionItem.Layer) && lockedOnTransitionAction)
                {
                    //SuperController.LogError(linkedActionItem.Name + " is not layered, skipping");
                    continue;
                }

                stringBuilder.AppendLine($"action: {linkedActionItem.Action}");
                stringBuilder.AppendLine($"when: {linkedActionItem.When}");
                if (linkedActionItem.Effect != null && linkedActionItem.Effect != "")
                {
                    stringBuilder.AppendLine($"effect: {linkedActionItem.Effect}");
                }
                if (!string.IsNullOrEmpty(linkedActionItem.Layer))
                {
                    stringBuilder.AppendLine($"layer: {linkedActionItem.Layer}");
                    stringBuilder.AppendLine("finalLayer: false");
                }
                stringBuilder.AppendLine();
            }
        }

        return stringBuilder.ToString().TrimEnd();
    }

    public static JSONStorable getPluginByType(Atom atom, string pluginType, bool optional = false)
    {
        foreach (string pluginId in atom.GetStorableIDs())
        {
            if (pluginId.Contains(pluginType))
            {
                return atom.GetStorableByID(pluginId);
            }
        }

        if (optional)
        {
            return null;
        }

        SuperController.LogError("Plugin with type " + pluginType + " has not been found on the atom " + atom.name);
        throw new NullReferenceException("Cannot find plugin of type " + pluginType + " on atom " + atom.uid);
    }

    public void Update()
    {
        try
        {
            ActionItem currentActionItem = GetItemByName(actionsChooser.val);
            if (currentActionItem != null && currentActionItem.Trigger.HasActionsWithActiveTimers)
            {
                currentActionItem.Trigger.Update();
            }
            foreach (var trigger in new List<Trigger>(activeTriggers))
            {
                if (trigger.HasActionsWithActiveTimers)
                {
                    trigger.Update();
                }
                else
                {
                    activeTriggers.Remove(trigger);
                }
            }
        }
        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
    }

    public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
    {
        base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
        actionsChooser.choices = actionsDictionary.Keys.ToList();
        actionsChooser.valNoCallback = actionsDictionary.Keys.FirstOrDefault();
        lastAction = GetItemByName(actionsChooser.val);
        RefreshUI();
    }

    private void RefreshUI()
    {
        if (!ShouldDeferRefresh())
        {
            Utils.RemoveUIElements(this, uiElements.Cast<object>().ToList());
            uiElements.Clear();
            if (!lockedOnTransitionAction)
            {
                actionsChooser.choices = actionsDictionary.Keys.ToList();
                actionsChooser.valNoCallback = actionsDictionary.Keys.FirstOrDefault();

                uiElements.Add(CreateScrollablePopup(actionsChooser, false));

                var executeButton = Utils.SetupButton(this, "Test Execute Action", OnExecuteActionButtonClicked, false);
                executeButton.buttonColor = new Color(0.45f, 1f, 0.45f);
                uiElements.Add(executeButton);

                uiElements.Add(Utils.SetupButton(this, "Add Action", OnAddActionButtonClicked, false));

                uiElements.Add(Utils.SetupButton(this, "Duplicate Action", OnDuplicateActionButtonClicked, false));

                var removeButton = Utils.SetupButton(this, "Remove Action", OnRemoveActionButtonClicked, false);
                removeButton.buttonColor = new Color(0.903f, 0.440f, 0.440f, 1.0f);
                uiElements.Add(removeButton);

                if (lastAction != null)
                {
                    actionsChooser.val = lastAction.Name;
                }
            }
            else
            {
                uiElements.Add(Utils.CreateTitle(this, "Blocked on transition action", false, 24, true, true));
                uiElements.Add(Utils.CreateTitle(this, "only layered actions can be executed", false, 24, true, true));
                uiElements.Add(Utils.CreateTitle(this, "waiting for TransitionActionFinished trigger", false, 24, true, true));
                var executeButton = Utils.SetupButton(this, "Force Unlock", OnTransitionActionFinished, false);
                executeButton.buttonColor = new Color(0.45f, 1f, 0.45f);
                uiElements.Add(executeButton);
            }

            RefreshActionUI();
            newLinkedActionsUI.RefreshUI();
        }
    }

    private bool ShouldDeferRefresh()
    {
        if (PluginUIEventsListener == null || PluginUIEventsListener.IsEnabled)
        {
            return false;
        }

        scheduleRefreshUI = true;
        return true;
    }

    public void OnDestroy()
    {
        DestroyImmediate(PluginUIEventsListener);
        PluginUIEventsListener = null;
        Utils.OnDestroyUI();
        UnityEngine.Object.Destroy(newLinkedActionsUI.backDrop);
    }

    public override bool ShouldIgnore() => false;
}
