using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using UnityEngine.Events;

using System.Collections;
using System;

public class LinkedActionsUI : UIPart
{
    public bool showAdvancedSettings = false;
    private ActionTextFieldsUI actionTextFieldsUI;
    public JSONStorableBool sameActionNameFunction;
    public JSONStorableString layerField;
    public JSONStorableBool transitionActionField;
    public JSONStorableBool playTimelineAnimField;
    private JSONStorableStringChooser actionsChooser;
    private Dictionary<string, ActionItem> actionsDictionary;

    public LinkedActionsUI(ActionListManager script) : base(script) { }

    public void Init(ActionTextFieldsUI actionTextFieldsUI, JSONStorableStringChooser actionChooser, Dictionary<string, ActionItem> actionsDictionary)
    {
        this.actionsChooser = actionChooser;
        this.actionsDictionary = actionsDictionary;
        this.actionTextFieldsUI = actionTextFieldsUI;

        sameActionNameFunction = new JSONStorableBool("Same Function And Name", true, OnSameFunctionAndNameChanged);
        layerField = new JSONStorableString("Layer", "", OnLayerChanged);
        transitionActionField = new JSONStorableBool("Transition Action", false, OnTransitionActionChanged);
        playTimelineAnimField = new JSONStorableBool("Play Timeline Animation", false, OnPlayTimelineAnimationChanged);
    }
    override
    public void RefreshUI()
    {
        RemoveAllUI();

        if (!script.lockedOnTransitionAction)
        {
            string currentSelectedAction = actionsChooser.val;
            if (string.IsNullOrEmpty(currentSelectedAction))
            {
                return;
            }

            ActionItem currentlySelectedActionItem = actionsDictionary[currentSelectedAction];
            transitionActionField.val = currentlySelectedActionItem.TransitionAction;
            playTimelineAnimField.val = currentlySelectedActionItem.PlayTimelineAnim;
            uiElements.Add(Utils.SetupButton(script, showAdvancedSettings ? "↑    Advanced Settings    ↑" : "↓    Advanced Settings    ↓", OnShowAdvancedClicked, true));
            if (showAdvancedSettings)
            {
                uiElements.Add(Utils.CreateTitle(script, "Allows to define separate function name", true, 24, true, true));
                uiElements.Add(Utils.CreateTitle(script, "that is sent to LLM", true, 24, true, true));
                uiElements.Add(Utils.SetupToggle(script, sameActionNameFunction, true));
                uiElements.Add(Utils.CreateTitle(script, "Locks action execution with the exclusion", true, 24, true, true));
                uiElements.Add(Utils.CreateTitle(script, "of layered actions", true, 24, true, true));
                uiElements.Add(Utils.SetupToggle(script, transitionActionField, true));
                uiElements.Add(Utils.CreateTitle(script, "On execution plays a Timeline animation", true, 24, true, true));
                uiElements.Add(Utils.CreateTitle(script, "with name 'VA_' + Function Name", true, 24, true, true));
                uiElements.Add(Utils.SetupToggle(script, playTimelineAnimField, true));
                uiElements.Add(Utils.CreateTitle(script, "Link all actions in this group together", true, 24, true, true));
                uiElements.Add(Utils.SetupButton(script, "Link Group", OnLinkAllActionsInGroup, true));
            }


            var groupsWithActions = actionsDictionary.Values
                .GroupBy(a => a.Group)
                .OrderBy(g => g.Key)
                .ToList();

            if (currentlySelectedActionItem.LinkedActions.Count() == 0)
            {
                uiElements.Add(Utils.CreateTitle(script, "Layered actions can't have any links", true, 24, true, true));
                uiElements.Add(Utils.CreateTitle(script, "and are executed alongside normal ones", true, 24, true, true));
                new TextInputBuilder(script, uiElements, "Layer", layerField).RightSide().Build();
            }

            if (!string.IsNullOrEmpty(actionsChooser.val))
            {
                var selectedAction = actionsDictionary[actionsChooser.val];
                copyFieldValues(selectedAction);
            }
            else
            {
                resetAllFields();
            }
        }
    }

    private void OnLinkAllActionsInGroup()
    {
        ActionItem currentActionItem = script.GetItemByName(actionsChooser.val);
        string group = currentActionItem.Group;

        List<ActionItem> groupedActionItems = new List<ActionItem>();
        foreach (var actionItem in actionsDictionary.Values.ToArray())
        {
            if (actionItem.Group == group)
            {
                groupedActionItems.Add(actionItem);
            }
        }

        foreach (var actionItem in groupedActionItems)
        {
            foreach (var itemToAdd in groupedActionItems)
            {
                if (itemToAdd != actionItem && !actionItem.LinkedActions.Contains(itemToAdd.Name))
                {
                    LinkActionItemToName(actionItem, itemToAdd.Name);
                }
            }
        }

        showAdvancedSettings = false;
        RefreshUI();
    }

    public void ApplySameFunctionNameToggle(ActionItem actionItem)
    {
        sameActionNameFunction.val = actionItem.Name == actionItem.Action;
    }
    private void OnTransitionActionChanged(bool value)
    {
        string actionKey = actionsChooser.val;
        if (!string.IsNullOrEmpty(actionKey) && actionsDictionary.ContainsKey(actionKey))
        {
            actionsDictionary[actionKey].TransitionAction = value;
        }
    }
    private void OnPlayTimelineAnimationChanged(bool value)
    {
        string actionKey = actionsChooser.val;
        if (!string.IsNullOrEmpty(actionKey) && actionsDictionary.ContainsKey(actionKey))
        {
            actionsDictionary[actionKey].PlayTimelineAnim = value;
        }
    }
    private void OnSameFunctionAndNameChanged(bool value)
    {
        if (!value)
        {
            actionTextFieldsUI.OnNameFieldChanged(ActionNameUtils.HumanFormatActionName(actionsChooser.val));
        }
        else
        {
            actionsDictionary[actionsChooser.val].Action = ActionNameUtils.FunctionFormatActionName(actionsChooser.val);
            sameActionNameFunction.val = true;
        }

        actionTextFieldsUI.RefreshUI();
    }

    private void OnLayerChanged(string newValue)
    {
        string actionKey = actionsChooser.val;
        if (!string.IsNullOrEmpty(actionKey) && actionsDictionary.ContainsKey(actionKey))
        {
            actionsDictionary[actionsChooser.val].Layer = newValue;
        }
    }
    private void OnRemoveAllButtonClicked()
    {
        string currentSelectedAction = actionsChooser.val;
        if (string.IsNullOrEmpty(currentSelectedAction))
        {
            SuperController.LogError("No action is currently selected.");
            return;
        }

        var currentActionItem = actionsDictionary[currentSelectedAction];
        if (currentActionItem.LinkedActions.Any())
        {
            currentActionItem.LinkedActions.Clear();

            RefreshUI();
        }
    }

    public void LinkActionItemToName(ActionItem actionItem, string nameToAdd)
    {
        if (!ActionItemHasSameActionName(actionItem, nameToAdd))
        {
            if (!string.IsNullOrEmpty(actionItem.Layer))
            {
                actionItem.Layer = "";
            }
            actionItem.LinkedActions.Add(nameToAdd);
        }
        else
        {
            SuperController.LogError("Action with function name = " + actionsDictionary[nameToAdd].Action + " already linked. Cannot have duplicate function names in the same route");
        }
    }

    private bool ActionItemHasSameActionName(ActionItem actionItem, string nameToAdd)
    {
        foreach (string name in actionItem.LinkedActions)
        {
            if (actionsDictionary[name].Action == actionsDictionary[nameToAdd].Action)
            {
                return true;
            }
        }
        return false;
    }

    public void resetAllFields()
    {
        layerField.val = "";
    }

    public void copyFieldValues(ActionItem actionItem)
    {
        if (actionItem != null && layerField != null)
        {
            layerField.val = actionItem.Layer;
        }
    }

    private void OnShowAdvancedClicked()
    {
        showAdvancedSettings = !showAdvancedSettings;
        RefreshUI();
    }
}
